#include <stdlib.h>
#include <string.h>
#include "ocapi.h"
#include <lauxlib.h>
#include <stdbool.h>

#include "interchange.h"

static int component_list(lua_State* L) {
  size_t filter_length;
  const char* filter_string = luaL_optlstring(L, 1, NULL, &filter_length);
  bool exact = lua_toboolean(L, 2);
  /* latch Component List Buffer (this step is required so that the list
     doesn't change while we're massaging it) */
  asm volatile("CDP p3, 5, cr0, cr0, cr0, #0");
  size_t list_size;
  /* read Component List Size Register */
  asm volatile("MRC p3, 0, %0, cr2, cr3, #0":"=r"(list_size):);
  void* list_buffer = get_io_buffer(list_size);
  if(list_buffer == NULL)
    return luaL_error(L, "out of memory");
  /* store Component List Buffer */
  asm volatile("STC p3, cr3, [%0]"::"r"(list_buffer):"memory");
  const void* p = list_buffer;
  interchange_tag ictype;
  lua_newtable(L);
  while((ictype = ic_get_type(p)) != ICTAG_END) {
    p = ic_push(L, p);
    if((ictype = ic_get_type(p)) == ICTAG_END)
      return luaL_error(L, "malformed Component List Buffer");
    if(filter_string) {
      size_t gotten_length;
      const char* gotten_string = lua_tolstring(L, -1, &gotten_length);
      if(gotten_string == NULL)
        return luaL_error(L, "malformed Component List Buffer");
      if(exact
         ? (gotten_length != filter_length
            || memcmp(gotten_string, filter_string, gotten_length))
         /* don't handle inline NULs here because I'm LAZY */
         : !strstr(gotten_string, filter_string)) {
        p = ic_skip(p);
        continue;
      }
    }
    p = ic_push(L, p);
    lua_settable(L, -3);
  }
  return 1;
}

static int component_type(lua_State* L) {
  /* not the most efficient way to do this, but very simple */
  lua_settop(L, 1);
  if(!lua_isstring(L, 1)) return luaL_error(L, "need a string argument");
  lua_pushcfunction(L, component_list);
  lua_call(L, 0, 1);
  lua_insert(L, -2);
  lua_gettable(L, -2);
  return 1;
}

static int component_slot(lua_State* L) {
  return luaL_error(L, "not implemented yet");
}

static int component_methods(lua_State* L) {
  return luaL_error(L, "not implemented yet");
}

static int component_invoke(lua_State* L) {
  return luaL_error(L, "not implemented yet");
}

static int component_doc(lua_State* L) {
  return luaL_error(L, "not implemented yet");
}

static const luaL_Reg componentlib[] = {
  {"list", component_list},
  {"type", component_type},
  {"slot", component_slot},
  {"methods", component_methods},
  {"invoke", component_invoke},
  {"doc", component_doc},
  {NULL, NULL}
};

int ocapi_open_component(lua_State* L) {
  luaL_newlib(L, componentlib);
  return 1;
}
