#include "ocapi.h"
#include "lauxlib.h"
#include <math.h>
#include "uuid.h"
#include <stdbool.h>
#include "malloc_extensions.h"
#include "interchange.h"

extern int __heap_start;
extern unsigned int ram_bytes;

static int computer_realTime(lua_State* L) {
  long long millis;
  asm("MRRC p3, 0, %0, %H0, cr1":"=r"(millis):);
  lua_pushnumber(L, millis / 1000.0);
  return 1;
}

static int computer_uptime(lua_State* L) {
  long long ticks;
  asm("MRRC p3, 0, %0, %H0, cr1":"=r"(ticks):);
  lua_pushnumber(L, ticks / 20.0);
  return 1;
}

static int computer_address(lua_State* L) {
  struct uuid uuid;
  asm("STC p3, cr7, %0":"=m"(uuid):);
  lua_pushstring(L, uuid2staticstring(&uuid));
  return 1;
}

static int computer_tmpAddress(lua_State* L) {
  struct uuid uuid;
  bool valid;
  asm("STC p3, cr7, %0\n\t"
      "MOVNE %1, #0\n\t"
      "MOVEQ %1, #1":"=m"(uuid),"=r"(valid):);
  if(valid)
    lua_pushstring(L, uuid2staticstring(&uuid));
  else
    lua_pushnil(L);
  return 1;
}

static int computer_users(lua_State* L) {
  return luaL_error(L, "not implemented yet");
}

static int computer_addUser(lua_State* L) {
  return luaL_error(L, "not implemented yet");
}

static int computer_removeUser(lua_State* L) {
  return luaL_error(L, "not implemented yet");
}

struct battery_status {
  double energy, max;
};
static int computer_energy(lua_State* L) {
  struct battery_status status;
  bool infinite_power;
  asm("STC p3, cr9, %0\n\t"
      "MOVEQ %1, #1\n\t"
      "MOVNE %1, #0":"=m"(status),"=r"(infinite_power):);
  if(infinite_power)
    lua_pushnumber(L, INFINITY);
  else
    lua_pushnumber(L, status.energy);
  return 1;
}
static int computer_maxEnergy(lua_State* L) {
  struct battery_status status;
  asm("STC p3, cr9, %0":"=m"(status):);
  lua_pushnumber(L, status.max);
  return 1;
}

static int computer_freeMemory(lua_State* L) {
  lua_pushnumber(L, get_heap_free_bytes());
  return 1;
}

static int computer_totalMemory(lua_State* L) {
  lua_pushnumber(L, get_heap_size());
  return 1;
}

static int computer_pushSignal(lua_State* L) {
  return luaL_error(L, "not implemented yet");
}

static int computer_popSignal(lua_State* L) {
  double timeout = luaL_optnumber(L, 1, 0.0);  
  if(timeout > 0.0) {
    timeout = timeout*20+0.5;
    if(timeout > 2147483647.0) timeout = 2147483647.0;
    long long now;
    asm volatile("MRRC p3, 0, %0, %H0, cr2":"=r"(now):);
    // NOT long long!
    now = now + (long)floor(timeout*20+0.5);
    asm volatile("MCRR p3, 0, %0, %H0, cr2"::"r"(now));
    // this will return immediately if there's already a signal
  }
  size_t size;
  asm("MRC p3, 0, %0, cr2, cr0":"=r"(size):);
  if(size == 0) return 0;
  /* size was non-zero, there is *definitely* a queued signal */
  void* io_buffer = get_io_buffer(size);
  asm("STC p3, cr0, [%0]"::"r"(io_buffer):"memory");
  const void* p = io_buffer;
  interchange_tag ictype;
  int value_count = 0;
  while((ictype = ic_get_type(p)) != ICTAG_END) {
    p = ic_push(L, p);
    lua_checkstack(L, 1);
    ++value_count;
  }
  return value_count;
}

static const luaL_Reg computerlib[] = {
  {"realTime", computer_realTime},
  {"uptime", computer_uptime},
  {"address", computer_address},
  {"freeMemory", computer_freeMemory},
  {"totalMemory", computer_totalMemory},
  {"pushSignal", computer_pushSignal},
  {"popSignal", computer_popSignal},
  {"tmpAddress", computer_tmpAddress},
  {"users", computer_users},
  {"addUser", computer_addUser},
  {"removeUser", computer_removeUser},
  {"energy", computer_energy},
  {"maxEnergy", computer_maxEnergy},
  {NULL, NULL}
};

int ocapi_open_computer(lua_State* L) {
  luaL_newlib(L, computerlib);
  return 1;
}
