#include "ocapi.h"
#include "lauxlib.h"
#include "gametime.h"
#include <inttypes.h>
#include <string.h>

#define LUA_STRFTIMEOPTIONS { "aAbBcCdDeFhHIjmMnprRStTwxXyY%", "" }

/* next five routines ripped pretty much directly from loslib.c */

static void setfield(lua_State *L, const char *key, int value) {
  lua_pushinteger(L, value);
  lua_setfield(L, -2, key);
}

static void setboolfield(lua_State *L, const char *key, int value) {
  if(value < 0)  /* undefined? */
    return;  /* does not set field */
  lua_pushboolean(L, value);
  lua_setfield(L, -2, key);
}

/* dead code */
/*
static int getboolfield(lua_State *L, const char *key) {
  int res;
  lua_getfield(L, -1, key);
  res = lua_isnil(L, -1) ? -1 : lua_toboolean(L, -1);
  lua_pop(L, 1);
  return res;
}
*/

static int getfield (lua_State *L, const char *key, int d) {
  int res, isnum;
  lua_getfield(L, -1, key);
  res = (int)lua_tointegerx(L, -1, &isnum);
  if (!isnum) {
    if (d < 0)
      return luaL_error(L, "field " LUA_QS " missing in date table", key);
    res = d;
  }
  lua_pop(L, 1);
  return res;
}

static const char *checkoption(lua_State *L, const char *conv, char *buff) {
  static const char *const options[] = LUA_STRFTIMEOPTIONS;
  unsigned int i;
  for(i = 0; i < sizeof(options)/sizeof(options[0]); i += 2) {
    if(*conv != '\0' && strchr(options[i], *conv) != NULL) {
      buff[1] = *conv;
      if(*options[i + 1] == '\0') {  /* one-char conversion specifier? */
        buff[2] = '\0';  /* end buffer */
        return conv + 1;
      }
      else if(*(conv + 1) != '\0' &&
               strchr(options[i + 1], *(conv + 1)) != NULL) {
        buff[2] = *(conv + 1);  /* valid two-char conversion specifier */
        buff[3] = '\0';  /* end buffer */
        return conv + 2;
      }
    }
  }
  luaL_argerror(L, 1, lua_pushfstring(L, "invalid conversion specifier '%%%s'", conv));
  return conv;  /* to avoid warnings */
}

/* original */

static int os_clock(lua_State* L) {
  union {
    double d;
    uint64_t ull;
  } cputime;
  asm volatile("MRRC p3, 0, %0, %H0, cr3":"=r"(cputime.ull):);
  lua_pushnumber(L, cputime.d);
  return 1;
}

/* next two routines ripped somewhat indirectly from loslib.c */

static int os_date(lua_State* L) {
  const char* s = luaL_optstring(L, 1, "%c");
  gametime_t t = luaL_opt(L, (gametime_t)luaL_checknumber, 2, gametime(NULL));
  struct gametm tm;
  if(*s == '!') ++s; // ignore UTC
  struct gametm* result = gametime2tm(&t, &tm);
  if(result == NULL) lua_pushnil(L);
  else if(!strcmp(s, "*t")) {
    lua_createtable(L, 0, 9);
    setfield(L, "sec", tm.second);
    setfield(L, "min", tm.minute);
    setfield(L, "hour", tm.hour);
    setfield(L, "day", tm.day);
    setfield(L, "month", tm.month);
    setfield(L, "year", tm.year);
    setfield(L, "wday", tm.weekDay);
    setfield(L, "yday", tm.yearDay);
    setboolfield(L, "isdst", 0);
  }
  else {
    char cc[4];
    luaL_Buffer b;
    cc[0] = '%';
    luaL_buffinit(L, &b);
    while(*s) {
      if(*s != '%') /* no conversion specifier? */
        luaL_addchar(&b, *s++);
      else {
        size_t reslen;
        char buff[200]; /* should be big enough for any conversion result */
        s = checkoption(L, s + 1, cc);
        reslen = gamestrftime(buff, sizeof(buff), cc, &tm);
        luaL_addlstring(&b, buff, reslen);
      }
    }
    luaL_pushresult(&b);
  }
  return 1;
}

static int os_time(lua_State* L) {
  gametime_t t;
  if(lua_isnoneornil(L, 1)) /* called without args? */
    t = gametime(NULL); /* get current time */
  else {
    struct gametm ts;
    luaL_checktype(L, 1, LUA_TTABLE);
    lua_settop(L, 1); /* make sure table is at the top */
    ts.second = getfield(L, "sec", 0);
    ts.minute = getfield(L, "min", 0);
    ts.hour = getfield(L, "hour", 12);
    ts.day = getfield(L, "day", -1);
    ts.month = getfield(L, "month", -1);
    ts.year = getfield(L, "year", -1);
    t = mkgametime(&ts);
  }
  if(t == (gametime_t)(-1))
    lua_pushnil(L);
  else
    lua_pushnumber(L, (lua_Number)t);
  return 1;
}

static const luaL_Reg oslib[] = {
  {"clock", os_clock},
  {"date", os_date},
  {"time", os_time},
  {NULL, NULL}
};

int ocapi_open_os(lua_State* L) {
  luaL_newlib(L, oslib);
  return 1;
}
