        .code 32
        .text

        /* add 3 to every digit of rT > 3 */
        .macro bcd_fixup rT
        AND r5, r3, \rT
        AND r6, r3, \rT, LSL #1
        ORR r5, r6
        AND r6, r3, \rT, LSR #1
        AND r6, r5
        AND r5, r3, \rT, LSR #2
        ORR r5, r6
        ORR r5, r5, r5, LSR #1
        ADD \rT, r5
        .endm
        
        .global dabble_16bit
        .func dabble_16bit
dabble_16bit:
        /* r0 = binary in
           r1 = pointer to bcd out
           r2 = bcd out
           r3 = constant (bit 2 of every hex digit)
           r4 = loop iterations remaining
           r5 = scratch 1
           r6 = scratch 2 */
        PUSH {r4, r5, r6, lr}
        LDR r3, =0x22222222
        MOV r4, #16
        LSL r0, r0, #16
        MOV r2, #0
1:      bcd_fixup r2
        /* shift next bit out */
        LSLS r0, #1
        RRX r2, r2
        ROR r2, #30
        /* repeat until we're out of bits */
        SUBS r4, #1
        BNE 1b
        STR r2, [r1]
        POP {r4, r5, r6, pc}
        .endfunc

        .global dabble_32bit
        .func dabble_32bit
dabble_32bit:
        /* r0 = binary in
           r1 = pointer to bcd out
           r2 = bcd out 1
           r3 = constant (bit 2 of every hex digit)
           r4 = loop iterations remaining
           r5 = scratch 1
           r6 = scratch 2
           r7 = bcd out 2 */
        PUSH {r4, r5, r6, r7, lr}
        LDR r3, =0x22222222
        MOV r4, #32
        MOV r2, #0
        MOV r7, #0
1:      /* fixup BCD digits */
        bcd_fixup r2
        bcd_fixup r7
        /* shift next bit out */
        /*LSLS r0, #1
        RRXS r2, r2
        ROR r2, #30
        RRX r7, r7
        ROR r7, #30*/
        LSL r7, #1
        TST r2, #0x80000000
        ORRNE r7, #1
        LSL r2, #1
        TST r0, #0x80000000
        ORRNE r2, #1
        LSL r0, #1
        /* repeat until we're out of bits */
        SUBS r4, #1
        BNE 1b
        STR r2, [r1]
        STR r7, [r1, #4]
        POP {r4, r5, r6, r7, pc}
        .endfunc

        .global dabble_64bit
        .func dabble_64bit
dabble_64bit:
        /* r0 = binary in hi
           r1 = binary in lo
           r2 = pointer to bcd out
           r3 = constant (bit 2 of every hex digit)
           r4 = loop iterations remaining
           r5 = scratch 1
           r6 = scratch 2
           r7 = bcd out 1
           r8 = bcd out 2
           (r9 is platform-defined)
           r10 = bcd out 3
           r11 = bcd out 4 */
        PUSH {r4, r5, r6, r7, r8, r10, r11, lr}
        LDR r3, =0x22222222
        MOV r4, #64
        MOV r7, #0
        MOV r8, #0
        MOV r10, #0
        MOV r11, #0
1:      bcd_fixup r7
        bcd_fixup r8
        bcd_fixup r10
        bcd_fixup r11
        /* shift next bit out */
        LSLS r1, #1
        RRXS r0, r0
        ROR r0, #30
        RRXS r7, r7
        ROR r7, #30
        RRXS r8, r8
        ROR r8, #30
        RRXS r10, r10
        ROR r10, #30
        RRX r11, r11
        ROR r11, #30
        /* repeat until we're out of bits */
        SUBS r4, #1
        BNE 1b
        STR r7, [r2]
        STR r8, [r2, #4]
        STR r10, [r2, #8]
        STR r11, [r2, #12]
        POP {r4, r5, r6, r7, r8, r10, r11, pc}
        .endfunc
