#include "gametime.h"
#include "die.h"
#include <math.h>

/* I'm not positive all this math is correct, but the important thing is that
   it's consistent with OpenComputers
   which this file achieves by borrowing heavily from the OpenComputers source
*/

static const uint8_t normal_year_month_lengths[12] = {
  31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31
};
static const uint8_t leap_year_month_lengths[12] = {
  31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31
};

static const uint8_t* month_lengths_for_year(int year) {
  if(year % 4 == 0 && (year % 100 != 0 || year % 400 == 0))
    return leap_year_month_lengths;
  else
    return normal_year_month_lengths;
}

gametime_t gametime(gametime_t* timep) {
  gametime_t time;
  asm("MRRC p3, 0, %0, %H0, cr0":"=r"(time):);
  if(timep != NULL) *timep = time;
  return time;
}

struct gametm* gametime2tm(const gametime_t* timep, struct gametm* tm) {
  gametime_t time;
  if(timep != NULL) time = *timep;
  else asm("MRRC p3, 0, %0, %H0, cr0":"=r"(time):);
  time = time + 6000; // start day at 6 o'clock
  long long absolute_day = time / 24000LL;
  tm->weekDay = (int)((4 + absolute_day) % 7);
  tm->year = 1970 + (int)(absolute_day / 364.2425);
  tm->yearDay = tm->day = (int)fmod(absolute_day, 365.2425);
  const uint8_t* month_lengths = month_lengths_for_year(tm->year);
  tm->month = 0;
  while(tm->day >= month_lengths[tm->month]) {
    tm->day -= month_lengths[tm->month];
    ++tm->month;
  }
  ++tm->month; // 1-based month
  ++tm->day; // 1-based day of month
  ++tm->yearDay; // 1-based day of year
  tm->second = (int)((time % 24000) * 60 * 60 / 1000);
  tm->minute = tm->second/60;
  tm->hour = (tm->minute / 60) % 24;
  tm->second %= 60;
  tm->minute %= 60;
  return tm;
}

size_t gamestrftime(char* s, size_t max, const char* format,
                    const struct gametm* tm) {
  die("Not implemented yet");
}

gametime_t mkgametime(const struct gametm* tm) {
  if(tm->year < 1970 || tm->month < 1 || tm->month > 12) return -1;
  const uint8_t* month_lengths = month_lengths_for_year(tm->year);
  int days = (int)ceil((tm->year - 1970) * 365.2425);
  int month = tm->month - 1;
  while(month > 0)
    days += month_lengths[--month];
  days += tm->day;
  gametime_t result = tm->second + (tm->minute + (tm->hour - 1 + days * 24) * 60) * 60 - 6000;
  return result;
}
