#include "lua.h"
#include "lualib.h"
#include "lauxlib.h"
#include "die.h"

#include <stdlib.h>

#include "ocapi.h"

/* based *very* loosely on Lua's standard interpreter */

static const luaL_Reg loadedlibs[] = {
  /* omit io, os, and package */
  {"_G", luaopen_base},
  {LUA_COLIBNAME, luaopen_coroutine},
  {LUA_TABLIBNAME, luaopen_table},
  {LUA_STRLIBNAME, luaopen_string},
  {LUA_BITLIBNAME, luaopen_bit32},
  {LUA_MATHLIBNAME, luaopen_math},
  {LUA_DBLIBNAME, luaopen_debug},
  /* OpenComputers-specific APIs */
  {"component", ocapi_open_component},
  {"computer", ocapi_open_computer},
  {"os", ocapi_open_os},
  {"system", ocapi_open_system},
  {"unicode", ocapi_open_unicode},
  /* Not needed for us? */
  //{"userdata", ocapi_open_userdata},
  {NULL, NULL}
};

static int pmain(lua_State* L) {
  /* very little garbage will be generated during initialization,
     suspend garbage collection */
  lua_gc(L, LUA_GCSTOP, 0);
  for(const luaL_Reg* lib = loadedlibs; lib->func; ++lib) {
    luaL_requiref(L, lib->name, lib->func, 1);
    lua_pop(L, 1);
  }
  lua_gc(L, LUA_GCRESTART, 0);
  return 0;
}

int main(int argc, char* argv[]) {
  lua_State* L = luaL_newstate();
  if(L == NULL) die("not enough memory");
  lua_pushcfunction(L, &pmain);
  switch(lua_pcall(L, 0, 0, 0)) {
  case LUA_OK: die("pmain returned");
  case LUA_ERRMEM: die("out of memory");
  case LUA_ERRERR: die("impossible!");
  case LUA_ERRGCMM:
  case LUA_ERRRUN:
    die(lua_tostring(L, -1));
  }
  __builtin_unreachable();
}
