#define _GLUE(a,b) a##b
#define GLUE(a,b) _GLUE(a,b)
static char* GLUE(out_sint, LENGTH_SPECIFIER)
     (char* p, ARG_TYPE _value, int width, int precision,
      bool left_justify, bool always_sign, bool always_sign_with_space,
      bool zeropad) {
  if(precision == DEFAULT_PRECISION) precision = 1;
  unsigned ARG_TYPE value;
  bool negative;
  if(_value < 0) { value = ~(unsigned ARG_TYPE)_value + 1; negative = true; }
  else { value = (unsigned ARG_TYPE)_value; negative = false; }
  unsigned int bcd_form[BCD_WORDS];
#if DABBLE_WIDTH == 16
  dabble_16bit(value, bcd_form);
#elif DABBLE_WIDTH == 32
  dabble_32bit(value, bcd_form);
#elif DABBLE_WIDTH == 64
  dabble_64bit(value, bcd_form);
#else
#error DABBLE_WIDTH must be 16, 32, or 64
#endif
  int digits = count_bcd_digits(bcd_form, BCD_WORDS);
  if(digits < precision) digits = precision;
  int actual_width = digits;
  if(always_sign || always_sign_with_space || negative)
    ++actual_width;
  if(!zeropad && !left_justify) p = pad(p, actual_width, width, ' ');
  if(negative) *p++ = '-';
  else if(always_sign) *p++ = '+';
  else if(always_sign_with_space) *p++ = ' ';
  if(zeropad) p = pad(p, actual_width, width, '0');
  p = output_bcd(p, digits, bcd_form, BCD_WORDS);
  if(!zeropad && left_justify) p = pad(p, actual_width, width, ' ');
  return p;
}

static char* GLUE(out_uint, LENGTH_SPECIFIER)
     (char* p, unsigned ARG_TYPE value, int width, int precision,
      bool left_justify, bool always_sign, bool always_sign_with_space,
      bool zeropad) {
  if(precision == DEFAULT_PRECISION) precision = 1;
  unsigned int bcd_form[BCD_WORDS];
#if DABBLE_WIDTH == 16
  dabble_16bit(value, bcd_form);
#elif DABBLE_WIDTH == 32
  dabble_32bit(value, bcd_form);
#elif DABBLE_WIDTH == 64
  dabble_64bit(value, bcd_form);
#else
#error DABBLE_WIDTH must be 16, 32, or 64
#endif
  int digits = count_bcd_digits(bcd_form, BCD_WORDS);
  if(digits < precision) digits = precision;
  int actual_width = digits;
  if(always_sign || always_sign_with_space)
    ++actual_width;
  if(!zeropad && !left_justify) p = pad(p, actual_width, width, ' ');
  if(always_sign) *p++ = '+';
  else if(always_sign_with_space) *p++ = ' ';
  if(zeropad) p = pad(p, actual_width, width, '0');
  p = output_bcd(p, digits, bcd_form, BCD_WORDS);
  if(!zeropad && left_justify) p = pad(p, actual_width, width, ' ');
  return p;
}

static char* GLUE(out_oint, LENGTH_SPECIFIER)
     (char* p, unsigned ARG_TYPE value, int width, int precision,
      bool left_justify, bool always_sign, bool always_sign_with_space,
      bool zeropad, bool alternate_form) {
  if(precision == DEFAULT_PRECISION) precision = 1;
  int digits = MAX_OCTAL_DIGITS;
  int shift = (MAX_OCTAL_DIGITS-1)*3;
  while(shift >= 0) {
    if(value >> shift) break;
    --digits;
    shift -= 3;
  }
  if(digits < precision) digits = precision;
  if(alternate_form && digits == precision) ++digits;
  int actual_width = digits;
  if(always_sign || always_sign_with_space)
    ++actual_width;
  if(!zeropad && !left_justify) p = pad(p, actual_width, width, ' ');
  if(always_sign) *p++ = '+';
  else if(always_sign_with_space) *p++ = ' ';
  if(zeropad) p = pad(p, actual_width, width, '0');
  shift = (digits-1)*3;
  while(shift >= 0) {
    *p++ = '0' + ((value>>shift)&7);
    shift -= 3;
  }
  if(!zeropad && left_justify) p = pad(p, actual_width, width, ' ');
  return p;
}

static char* GLUE(out_xint, LENGTH_SPECIFIER)
     (char* p, unsigned ARG_TYPE value, const char* digitchars,
      int width, int precision, bool left_justify, bool always_sign,
      bool always_sign_with_space, bool zeropad, bool alternate_form) {
  if(precision == DEFAULT_PRECISION) precision = 1;
  int digits = MAX_HEX_DIGITS;
  int shift = (MAX_HEX_DIGITS-1)*4;
  while(shift >= 0) {
    if(value >> shift) break;
    --digits;
    shift -= 4;
  }
  if(digits < precision) digits = precision;
  int actual_width = digits;
  if(always_sign || always_sign_with_space)
    ++actual_width;
  if(alternate_form) actual_width += 2;
  if(!zeropad && !left_justify) p = pad(p, actual_width, width, ' ');
  if(always_sign) *p++ = '+';
  else if(always_sign_with_space) *p++ = ' ';
  if(alternate_form) {
    *p++ = '0';
    *p++ = 'x';
  }
  if(zeropad) p = pad(p, actual_width, width, '0');
  shift = (digits-1)*4;
  while(shift >= 0) {
    *p++ = digitchars[(value>>shift)&15];
    shift -= 4;
  }
  if(!zeropad && left_justify) p = pad(p, actual_width, width, ' ');
  return p;
}
