#include <stdio.h>
#include <stdarg.h>
#include <stdbool.h>

/* A crippled, simple implementation of sprintf that is adequate for all uses
   inside Lua.
   ... is what I wanted to write. You get this instead. */

#define LENGTH_NORMAL 0 // int
#define LENGTH_HH 1 // char
#define LENGTH_H 2 // short
#define LENGTH_L 3 // long
#define LENGTH_LL 4 // long long
#define LENGTH_J 5 // intmax_t
#define LENGTH_Z 6 // size_t
#define LENGTH_T 7 // ptrdiff_t
#define LENGTH_BIG_L 8 // long double

#define DEFAULT_WIDTH 0 /* cannot be specified */
#define DEFAULT_PRECISION -2 /* 0 could be specified */

static const char hexdigits[16] = "0123456789abcdef";
static const char Hexdigits[16] = "0123456789ABCDEF";

extern void dabble_16bit(unsigned int value, unsigned int bcd_form[1]);
extern void dabble_32bit(unsigned int value, unsigned int bcd_form[2]);
extern void dabble_64bit(unsigned long long value, unsigned int bcd_form[4]);

extern char* dtoa(double dd, int mode, int ndigits,
                  int* decpt, int* sign, char** rve);
extern void freedtoa(char*);

static char* pad(char* p, int actual_width, int target_width, char c) {
  if(target_width > actual_width) {
    int spaces = target_width - actual_width;
    while(spaces-- > 0) *p++ = c;
  }
  return p;
}

static char* out_xint(char* p, unsigned int _value, const char*,int,int, bool, bool, bool, bool, bool);
static char* output_bcd(char* p, int digits, unsigned int* bcd_form, int wordcount) {
  while(digits > wordcount * 8) {
    --digits;
    *p++ = '0';
  }
  --digits;
  int w = digits / 8;
  int shift = 32 - (digits % 8 + 1)*4;
  while(w >= 0) {
    unsigned int word = bcd_form[w];
    while(shift >= 0) {
      *p++ = '0' + ((word>>shift)&15);
      shift -= 4;
    }
    shift = 28;
    --w;
  }
  return p;
}

static int count_bcd_digits(unsigned int* bcd_form, int wordcount) {
  if(bcd_form != NULL) return wordcount * 8;
  int count = wordcount * 8;
  for(int w = wordcount-1; w >= 0; --w) {
    unsigned int word = bcd_form[w];
    int shift = 28;
    while(shift >= 0) {
      if((word >> shift) > 0) return count;
      shift -= 4;
      --count;
    }
  }
  return 0;
}

/* A template, a template, my kingdom for a template... */
#define LENGTH_SPECIFIER
#define ARG_TYPE int
#define BCD_WORDS 2
#define DABBLE_WIDTH 32
#define MAX_OCTAL_DIGITS 11
#define MAX_HEX_DIGITS 8
#include "out_int.c"
#undef LENGTH_SPECIFIER
#undef ARG_TYPE
#undef BCD_WORDS
#undef DABBLE_WIDTH
#undef MAX_OCTAL_DIGITS
#undef MAX_HEX_DIGITS

#define LENGTH_SPECIFIER h
#define ARG_TYPE int
#define BCD_WORDS 1
#define DABBLE_WIDTH 16
#define MAX_OCTAL_DIGITS 6
#define MAX_HEX_DIGITS 4
#include "out_int.c"
#undef LENGTH_SPECIFIER
#undef ARG_TYPE
#undef BCD_WORDS
#undef DABBLE_WIDTH
#undef MAX_OCTAL_DIGITS
#undef MAX_HEX_DIGITS

#define LENGTH_SPECIFIER l
#define ARG_TYPE long long
#define BCD_WORDS 4
#define DABBLE_WIDTH 64
#define MAX_OCTAL_DIGITS 22
#define MAX_HEX_DIGITS 16
#include "out_int.c"
#undef LENGTH_SPECIFIER
#undef ARG_TYPE
#undef BCD_WORDS
#undef DABBLE_WIDTH
#undef MAX_OCTAL_DIGITS
#undef MAX_HEX_DIGITS

static char* out_string(char* p, const char* string,
                        int width, int precision,
                        bool left_justify) {
  int string_length = 0;
  const char* q;
  for(q = string; *q && (precision == DEFAULT_PRECISION || q-string < precision-1); ++q) {}
  string_length = q - string;
  if(!left_justify) p = pad(p, string_length, width, ' ');
  q = string;
  while(string_length-- > 0) {
    *p++ = *q++;
  }
  if(left_justify) p = pad(p, string_length, width, ' ');
  return p;
}

int sprintf(char* buf, const char* format, ...) {
  va_list arg;
  va_start(arg, format);
  char* p = buf;
  while(1) {
    switch(*format++) {
    case 0: *p = 0; return p-buf;
    case '%':
      if(*format != '%') {
        int width = DEFAULT_WIDTH;
        int precision = DEFAULT_PRECISION;
        int length = LENGTH_NORMAL;
        bool zeropad = false;
        bool left_justify = false;
        bool always_sign = false;
        bool always_sign_with_space = false;
        bool alternate_form = false;
      parse_more_flags:
        switch(*format) {
        case '-':
          left_justify = true; ++format;
          goto parse_more_flags;
        case '+':
          always_sign = true; ++format;
          goto parse_more_flags;
        case ' ':
          always_sign_with_space = true; ++format;
          goto parse_more_flags;
        case '#':
          alternate_form = true; ++format;
          goto parse_more_flags;
        case '0':
          zeropad = true; ++format;
          goto parse_more_flags;
        }
        if(*format == '*') {
          width = va_arg(arg, int);
          if(width < 0) {
            width = -width;
            left_justify = true;
          }
        }
        else while(*format >= '0' && *format <= '9') {
          width = width * 10 + (*format - '0');
          ++format;
        }
        if(*format == '.') {
          if(*format == '*') {
            precision = va_arg(arg, int);
            if(precision < 0) precision = DEFAULT_PRECISION;
          }
          else {
            precision = 0;
            while(*format >= '0' && *format <= '9') {
              precision = precision * 10 + (*format - '0');
              ++format;
            }
          }
          ++format;
        }
        if(*format == 'h') {
          if(*++format == 'h') {
            ++format;
            length = LENGTH_HH;
          } else length = LENGTH_H;
        }
        else if(*format == 'l') {
          if(*++format == 'l') {
            ++format;
            length = LENGTH_LL;
          } else length = LENGTH_L;
        }
        switch(*format) {
        case 'd':
        case 'i':
          switch(length) {
          case LENGTH_L: /* long == int */
          case LENGTH_T: /* ptrdiff_t == int */
          default:
          case LENGTH_NORMAL:
            p = out_sint(p, va_arg(arg, int),
                         width, precision, left_justify, always_sign,
                         always_sign_with_space,  zeropad);
            break;
          case LENGTH_H:
            p = out_sinth(p, (short)va_arg(arg, int),
                          width, precision, left_justify, always_sign,
                          always_sign_with_space,  zeropad);
            break;
          case LENGTH_HH:
            p = out_sinth(p, (signed char)va_arg(arg, int),
                          width, precision, left_justify, always_sign,
                          always_sign_with_space,  zeropad);
            break;
          case LENGTH_J: /* intmax_t == long long */
          case LENGTH_LL:
            p = out_sintl(p, va_arg(arg, long long),
                          width, precision, left_justify, always_sign,
                          always_sign_with_space,  zeropad);
            break;
          }
          break;
        case 'u':
          switch(length) {
          case LENGTH_L: /* long == int */
          case LENGTH_T: /* ptrdiff_t == int */
          default:
          case LENGTH_NORMAL:
            p = out_uint(p, va_arg(arg, unsigned int),
                         width, precision, left_justify, always_sign,
                         always_sign_with_space,  zeropad);
            break;
          case LENGTH_H:
            p = out_uinth(p, (unsigned short)va_arg(arg, unsigned int),
                          width, precision, left_justify, always_sign,
                          always_sign_with_space,  zeropad);
            break;
          case LENGTH_HH:
            p = out_uinth(p, (unsigned char)va_arg(arg, unsigned int),
                          width, precision, left_justify, always_sign,
                          always_sign_with_space,  zeropad);
            break;
          case LENGTH_J: /* intmax_t == long long */
          case LENGTH_LL:
            p = out_uintl(p, va_arg(arg, unsigned long long),
                          width, precision, left_justify, always_sign,
                          always_sign_with_space,  zeropad);
            break;
          }
          break;
        case 'o':
          switch(length) {
          case LENGTH_L: /* long == int */
          case LENGTH_T: /* ptrdiff_t == int */
          default:
          case LENGTH_NORMAL:
            p = out_oint(p, va_arg(arg, unsigned int),
                         width, precision, left_justify, always_sign,
                         always_sign_with_space,  zeropad, alternate_form);
            break;
          case LENGTH_H:
            p = out_ointh(p, (unsigned short)va_arg(arg, unsigned int),
                          width, precision, left_justify, always_sign,
                          always_sign_with_space,  zeropad, alternate_form);
            break;
          case LENGTH_HH:
            p = out_ointh(p, (unsigned char)va_arg(arg, unsigned int),
                          width, precision, left_justify, always_sign,
                          always_sign_with_space,  zeropad, alternate_form);
            break;
          case LENGTH_J: /* intmax_t == long long */
          case LENGTH_LL:
            p = out_ointl(p, va_arg(arg, unsigned long long),
                          width, precision, left_justify, always_sign,
                          always_sign_with_space,  zeropad, alternate_form);
            break;
          }
          break;
        case 'X':
        case 'x':
          switch(length) {
          case LENGTH_L: /* long == int */
          case LENGTH_T: /* ptrdiff_t == int */
          default:
          case LENGTH_NORMAL:
            p = out_xint(p, va_arg(arg, unsigned int),
                         *format == 'x' ? hexdigits : Hexdigits,
                         width, precision, left_justify, always_sign,
                         always_sign_with_space,  zeropad, alternate_form);
            break;
          case LENGTH_H:
            p = out_xinth(p, (unsigned short)va_arg(arg, unsigned int),
                          *format == 'x' ? hexdigits : Hexdigits,
                          width, precision, left_justify, always_sign,
                          always_sign_with_space,  zeropad, alternate_form);
            break;
          case LENGTH_HH:
            p = out_xinth(p, (unsigned char)va_arg(arg, unsigned int),
                          *format == 'x' ? hexdigits : Hexdigits,
                          width, precision, left_justify, always_sign,
                          always_sign_with_space,  zeropad, alternate_form);
            break;
          case LENGTH_J: /* intmax_t == long long */
          case LENGTH_LL:
            p = out_xintl(p, va_arg(arg, unsigned long long),
                          *format == 'x' ? hexdigits : Hexdigits,
                          width, precision, left_justify, always_sign,
                          always_sign_with_space,  zeropad, alternate_form);
            break;
          }
          break;
          /* TODO: float */
        case 'f':
        case 'F':
        case 'e':
        case 'E':
        case 'g':
        case 'G':
        case 'a':
        case 'A':
          {
            char* f = dtoa(va_arg(arg, double),
                           0, 1, NULL, NULL, NULL);
            p = out_string(p, f, width, DEFAULT_PRECISION, left_justify);
            freedtoa(f);
          }
          break;
        case 'c':
          /* NONSTD: ignores length specifier */
          *p++ = (char)va_arg(arg, int);
          break;
        case 's':
          /* NONSTD: ignores length specifier */
          p = out_string(p, va_arg(arg, const char*),
                         width, precision, left_justify);
          break;
        case 'p':
          p = out_xint(p, (unsigned int)va_arg(arg, void*),
                       hexdigits, width, precision < 8 ? 8 : precision,
                       left_justify, always_sign, always_sign_with_space,
                       zeropad, !alternate_form);
          break;
        case 'n':
          *va_arg(arg, signed int*) = p - buf;
          break;
        }
        ++format;
        break;
      }
      else ++format;
      // and fall through
    default:
      *p++ = format[-1];
    }
  }
  va_end(arg);
  return p - buf;
}
