#include <string.h>
#include <inttypes.h>
#include <stdbool.h>

const char* strchr(const char* s, int c) {
  do {
    if(*s == c) return s;
    else if(*s == 0) return NULL;
    ++s;
  } while(1);
  __builtin_unreachable();
}

const char* strrchr(const char* s, int c) {
  const char* ret = NULL, *nu = s;
  while((nu = strchr(nu, c)) != NULL)
    ret = nu;
  return ret;
}

const char* strpbrk(const char* s, const char* accept) {
  while(*s) {
    const char* p = accept;
    while(*p) if(*s == *p++) return s;
    ++s;
  }
  return NULL;
}

size_t strlen(const char* s) {
  size_t ret = 0;
  while(*s++) ++ret;
  return ret;
}

int strcmp(const char* a, const char* b) {
  while(*a || *b) {
    if(*a > *b) return 1;
    else if(*a < *b) return -1;
    ++a; ++b;
  }
  return 0;
}

const char* strstr(const char* haystack, const char* needle) {
  while(*haystack) {
    if(!strcmp(haystack, needle)) return haystack;
  }
  return NULL;
}

/* TODO: Naive. Slow. Replace. */
void* memcpy(void* _dst, const void* _src, size_t n) {
  uint8_t* dst = (uint8_t*)_dst;
  const uint8_t* src = (const uint8_t*)_src;
  while(n-- > 0)
    *dst++ = *src++;
  return _dst;
}

/* TODO: Naive. Slow. Replace. */
void* memmove(void* _dst, const void* _src, size_t n) {
  uint8_t* dst = (uint8_t*)_dst;
  const uint8_t* src = (const uint8_t*)_src;
  if(src >= dst || dst - src >= n) return memcpy(_dst, _src, n);
  else {
    dst = dst + n - 1;
    src = src + n - 1;
    while(n-- > 0)
      *dst-- = *src--;
    return _dst;
  }
}

/* TODO: Naive. Slow. Replace. */
int memcmp(const void* _a, const void* _b, size_t n) {
  const uint8_t* a = (const uint8_t*)_a;
  const uint8_t* b = (const uint8_t*)_b;
  while(n-- > 0) {
    if(*a > *b) return 1;
    else if(*a < *b) return -1;
    ++a; ++b;
  }
  return 0;
}

/* TODO: Naive. Slow. Replace. */
void* memset(void* ret, int c, size_t n) {
  uint8_t* p = (uint8_t*)ret;
  while(n-- > 0) *p++ = c;
  return ret;
}

char* strcpy(char* dst, const char* src) {
  char* ret = dst;
  while((*dst++ = *src++))
    ;
  return ret;
}

size_t strspn(const char* s, const char* accept) {
  size_t ret = 0;
  while(*s) {
    const char* p = accept;
    bool accepted = false;
    while(*p) {
      if(*s == *p) { accepted = true; break; }
      ++p;
    }
    if(!accepted) return ret;
    ++ret;
    ++s;
  }
  return ret;
}

void* memchr(const void* _s, int c, size_t n) {
  const uint8_t* s = (const uint8_t*)_s;
  while(n-- > 0) {
    if(*s == c) return (void*)s;
  }
  return NULL;
}
